from .base import *
from .deepLC import *
from .prosit import *

__all__ = [
    "RetentionTimePredictor",
    "PrositRetentionTimePredictor",
    "DeepLCRetentionTimePredictor",
    "PrositIntensityPredictor",
]



import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras import regularizers, constraints, initializers, activations


class DecoderAttentionLayer(tf.keras.layers.Layer):
    def __init__(self, time_steps):
        super(DecoderAttentionLayer, self).__init__()
        self.time_steps = time_steps

    def build(self, input_shape):
        self.permute = tf.keras.layers.Permute((2, 1))
        self.dense = tf.keras.layers.Dense(self.time_steps, activation="softmax")
        self.multiply = tf.keras.layers.Multiply()

    def call(self, inputs):
        x = self.permute(inputs)
        x = self.dense(x)
        x = self.permute(x)
        x = self.multiply([inputs, x])
        return x


class AttentionLayer(tf.keras.layers.Layer):
    def __init__(
        self,
        context=False,
        W_regularizer=None,
        b_regularizer=None,
        u_regularizer=None,
        W_constraint=None,
        b_constraint=None,
        u_constraint=None,
        bias=True,
        **kwargs
    ):

        self.supports_masking = True
        self.init = initializers.get("glorot_uniform")
        self.W_regularizer = regularizers.get(W_regularizer)
        self.b_regularizer = regularizers.get(b_regularizer)
        self.u_regularizer = regularizers.get(u_regularizer)
        self.W_constraint = constraints.get(W_constraint)
        self.b_constraint = constraints.get(b_constraint)
        self.u_constraint = constraints.get(u_constraint)
        self.bias = bias
        self.context = context
        super(AttentionLayer, self).__init__(**kwargs)

    def build(self, input_shape):
        assert len(input_shape) == 3
        self.W = self.add_weight(
            shape=(input_shape[-1],),
            initializer=self.init,
            name="{}_W".format(self.name),
            regularizer=self.W_regularizer,
            constraint=self.W_constraint,
        )
        if self.bias:
            self.b = self.add_weight(
                shape=(input_shape[1],),
                initializer="zero",
                name="{}_b".format(self.name),
                regularizer=self.b_regularizer,
                constraint=self.b_constraint,
            )
        else:
            self.b = None
        if self.context:
            self.u = self.add_weight(
                shape=(input_shape[-1],),
                initializer=self.init,
                name="{}_u".format(self.name),
                regularizer=self.u_regularizer,
                constraint=self.u_constraint,
            )

        self.built = True

    def compute_mask(self, input, input_mask=None):
        return None

    def call(self, x, mask=None):
        a = K.squeeze(K.dot(x, K.expand_dims(self.W)), axis=-1)
        if self.bias:
            a += self.b
        a = K.tanh(a)
        if self.context:
            a = K.squeeze(K.dot(x, K.expand_dims(self.u)), axis=-1)
        a = K.exp(a)
        if mask is not None:
            a *= K.cast(mask, K.floatx())
        a /= K.cast(K.sum(a, axis=1, keepdims=True) + K.epsilon(), K.floatx())
        a = K.expand_dims(a)
        weighted_input = x * a
        return K.sum(weighted_input, axis=1)

    def compute_output_shape(self, input_shape):
        return input_shape[0], input_shape[-1]

    def get_config(self):
        config = {
            "bias": self.bias,
            "context": self.context,
            "W_regularizer": regularizers.serialize(self.W_regularizer),
            "b_regularizer": regularizers.serialize(self.b_regularizer),
            "u_regularizer": regularizers.serialize(self.u_regularizer),
            "W_constraint": constraints.serialize(self.W_constraint),
            "b_constraint": constraints.serialize(self.b_constraint),
            "u_constraint": constraints.serialize(self.u_constraint),
        }
        base_config = super(AttentionLayer, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



from os.path import join
from matplotlib import pyplot as plt
from dlomix.reports.Report import PDFFile, Report
import numpy as np
from matplotlib.colors import LogNorm
from matplotlib.ticker import LogLocator


class RetentionTimeReport(Report):
    """Report generation for Retention Time Prediction tasks."""

    TARGETS_LABEL = "iRT (measured)"
    PREDICTIONS_LABEL = "iRT (predicted)"

    def __init__(self, output_path, history, figures_ext="png"):
        super(RetentionTimeReport, self).__init__(output_path, history, figures_ext)

        self.pdf_file = PDFFile("DLOmix - Retention Time Report")

    def generate_report(self, targets, predictions, **kwargs):
        self._init_report_resources()

        r2 = self.calculate_r2(targets, predictions)
        self.plot_all_metrics()
        self.plot_residuals(targets, predictions)
        self.plot_density(targets, predictions)

        self._compile_report_resources_add_pdf_pages()

        self.pdf_file.output(join(self._output_path, "iRT_Report.pdf"), "F")

    def calculate_r2(self, targets, predictions):
        """Calculate R-squared using sklearn given true targets and predictions

        Arguments
        ---------
            targets: Array with target values
            predictions: Array with prediction values

        Returns:
            r_squared (float): float value of R squared
        """
        from sklearn.metrics import r2_score

        r2 = r2_score(np.ravel(targets), np.ravel(predictions))

        self._add_report_resource(
            "r2",
            "R-Squared",
            f"The R-squared value for the predictions is {round(r2, 4)}",
            r2,
        )

        return r2

    def plot_residuals(self, targets, predictions, xrange=(0, 0)):
        """Plot histogram of residuals

        Argsuments
        ----------
            targets: Array with target values
            predictions: Array with prediction values
            xrange (tuple, optional): X-axis range for plotting the histogram. Defaults to (-10, 10).
        """
        error = np.ravel(targets) - np.ravel(predictions)

        x_min, x_max = xrange
        if xrange == (0, 0):
            mean, std_dev = np.mean(error), np.std(error)
            x_min, x_max = mean - (3 * std_dev), mean + (3 * std_dev)

        bins = np.linspace(x_min, x_max, 200)

        plt.hist(error, bins, alpha=0.5, color="orange")
        plt.title("Historgram of Residuals")
        plt.xlabel("Residual value")
        plt.ylabel("Count")
        save_path = join(self._output_path, "histogram_residuals" + self._figures_ext)
        plt.savefig(save_path)
        plt.show()
        plt.close()

        self._add_report_resource(
            "residuals_plot",
            "Error Residuals",
            "The following plot shows a historgram of residuals for the test data.",
            save_path,
        )

    def plot_density(
        self,
        targets,
        predictions,
        irt_delta95=5,
        palette="Reds_r",
        delta95_line_color="#36479E",
        nbins=1000,
    ):
        """Create density plot

        Arguments
        ---------
            targets:  Array with target values
            predictions:  Array with prediction values
            irt_delta95 (int, optional): iRT Value of the delta 95% . Defaults to 5.
            palette (str, optional): Color palette from matplotlib. Defaults to 'Reds_r'.
            delta95_line_color (str, optional): Color for the delta 95% line. Defaults to '#36479E'.
            nbins (int, optional): Number of bins to use for creating the 2D histogram. Defaults to 1000.
        """

        H, xedges, yedges = np.histogram2d(targets, predictions, bins=nbins)

        x_min = np.min(targets)
        x_max = np.max(targets)

        # H needs to be rotated and flipped
        H = np.rot90(H)
        H = np.flipud(H)

        # Mask zeros
        Hmasked = np.ma.masked_where(H == 0, H)  # Mask pixels with a value of zero

        # Plot 2D histogram using pcolor
        cm = plt.cm.get_cmap(palette)
        plt.pcolormesh(
            xedges, yedges, Hmasked, cmap=cm, norm=LogNorm(vmin=1e0, vmax=1e2)
        )

        plt.xlabel(RetentionTimeReport.TARGETS_LABEL, fontsize=18)
        plt.ylabel(RetentionTimeReport.PREDICTIONS_LABEL, fontsize=18)

        cbar = plt.colorbar(ticks=LogLocator(subs=range(5)))
        cbar.ax.set_ylabel("Counts", fontsize=14)

        plt.plot([x_min, x_max], [x_min, x_max], c="black")
        plt.plot(
            [x_min, x_max],
            [x_min - irt_delta95, x_max - irt_delta95],
            color=delta95_line_color,
        )
        plt.plot(
            [x_min, x_max],
            [x_min + irt_delta95, x_max + irt_delta95],
            color=delta95_line_color,
        )

        font_size = 14  # Adjust as appropriate.
        cbar.ax.tick_params(labelsize=font_size)
        cbar.ax.minorticks_on()
        save_path = join(self._output_path, "density_plot" + self._figures_ext)
        plt.savefig(save_path)
        plt.show()
        plt.close()

        self._add_report_resource(
            "density_plot",
            "Density Plot",
            "The following figure shows the density plot with the delta-95 highlighted for the test data.",
            save_path,
        )



DEFAULT_PARQUET_ENGINE = "pyarrow"

retention_time_pipeline_parameters = {
    "model_params": {"seq_length": 30},
    "data_params": {
        "seq_length": 30,
    },
    "trained_model_path": "../pretrained_models/retention_time/example_rtmodel/",
    "trained_model_zipfile_name": "rtmodel.zip",
    "trained_model_stats": [0.0, 1.0],
}

retention_time_pipeline_parameters.update(
    {
        "trained_model_url": "https://raw.githubusercontent.com/wilhelm-lab/dlomix/develop"
        + retention_time_pipeline_parameters["trained_model_path"].strip("..")
        + retention_time_pipeline_parameters["trained_model_zipfile_name"]
    }
)

ALPHABET_UNMOD = {
    "A": 1,
    "C": 2,
    "D": 3,
    "E": 4,
    "F": 5,
    "G": 6,
    "H": 7,
    "I": 8,
    "K": 9,
    "L": 10,
    "M": 11,
    "N": 12,
    "P": 13,
    "Q": 14,
    "R": 15,
    "S": 16,
    "T": 17,
    "V": 18,
    "W": 19,
    "Y": 20,
}



from dlomix.losses import masked_spectral_distance
import numpy as np
import tensorflow as tf

import logging

logger = logging.getLogger(__name__)



# ------------------ intensity - masked spectral distance ------------------

def test_spectral_distance_identical():
    y_true = [[0.1, 0.2, 0.3]]
    y_true_tensor = tf.convert_to_tensor(y_true)
    y_pred_tensor = tf.convert_to_tensor(y_true)

    sa = masked_spectral_distance(y_true_tensor, y_pred_tensor)
    logger.info("Spectral Angle for identical vectors: {}".format(sa.numpy()))

    assert sa.numpy() == 0

def test_spectral_distance_different():
    y_true = [[0.1, 0.2, 0.3]]
    y_true_tensor = tf.convert_to_tensor(y_true)
    y_pred_tensor = tf.convert_to_tensor([list(reversed(y_true[0]))])

    sa = masked_spectral_distance(y_true_tensor, y_pred_tensor)
    logger.info("Spectral Angle for reversed vectors: {}".format(sa.numpy()))

    assert sa.numpy() != 0

def test_spectral_distance_zero_input():
    y_true = [[0., 0., 0.]]
    y_true_tensor = tf.convert_to_tensor(y_true)
    y_pred_tensor = tf.convert_to_tensor(y_true)

    sa = masked_spectral_distance(y_true_tensor, y_pred_tensor)
    logger.info("Spectral Angle for zero input vectors: {}".format(sa.numpy()))

# ---------------------------------------------------------------------------


import numpy as np
from dlomix.constants import retention_time_pipeline_parameters
from dlomix.data.RetentionTimeDataset import RetentionTimeDataset
from dlomix.models.base import RetentionTimePredictor
from dlomix.reports import RetentionTimeReport
from os.path import join, dirname, splitext
from os import makedirs
import requests
import zipfile

# pipelines can be used to train the model further or from scratch given a dataset
# add string arguments (e.g. prosit to create the model, data source to create the dataset)

# if neither  train nor test are provided --> use toy datasets to (train if necessary or load pre-trained weights), predict on test, and generate report
# if test only --> load pre-trained weights, predict and generate report
# if train and test --> do what you have to do


class RetentionTimePipeline:
    def __init__(self, pre_trained=True):
        super(RetentionTimePipeline, self).__init__()
        self.model = None
        self.test_dataset = None
        self.pre_trained = pre_trained

        # pass the config in the constructor
        # refactor to have a base class Pipeline

        self._build_model()

    def _build_model(self):
        self.model = RetentionTimePredictor(
            **retention_time_pipeline_parameters["model_params"]
        )

        if self.pre_trained:
            self._download_unzip_pretrained_model(
                retention_time_pipeline_parameters["trained_model_url"],
                retention_time_pipeline_parameters["trained_model_path"]
                + retention_time_pipeline_parameters["trained_model_zipfile_name"],
            )

            self.model.load_weights(
                retention_time_pipeline_parameters["trained_model_path"]
                + splitext(
                    retention_time_pipeline_parameters["trained_model_zipfile_name"]
                )[0]
            )

    def _download_unzip_pretrained_model(self, model_url, save_path):
        makedirs(model_url)
        r = requests.get(model_url)

        with open(save_path, "wb") as f:
            f.write(r.content)

        self._unzip_model(save_path)

    def _unzip_model(self, model_zipfile_path):
        zip_ref = zipfile.ZipFile(model_zipfile_path)
        model_folder = dirname(model_zipfile_path)
        zip_ref.extractall(model_folder)
        zip_ref.close()

    """
    
    Predict retention times given data either as numpy array of sequences or a filepath to a csv file
    
    """

    def predict(self, data=None):
        if not (isinstance(data, str) or isinstance(data, np.ndarray)):
            raise ValueError(
                "Dataset should be provided either as a numpy array or a string pointing to a file."
            )

        self.test_dataset = RetentionTimeDataset(
            data,
            **retention_time_pipeline_parameters["data_params"],
            val_ratio=0,
            test=True
        )
        (
            self.test_dataset.data_mean,
            self.test_dataset.data_std,
        ) = retention_time_pipeline_parameters["trained_model_stats"]

        predictions = self.model.predict(self.test_dataset.test_data)
        predictions = self.test_dataset.denormalize_targets(predictions)
        predictions = predictions.ravel()

        return predictions

    def predict_report(self, data, output_path="./") -> None:
        predictions = self.predict(data)
        report = RetentionTimeReport(output_path=output_path, history=None)

        test_targets = self.test_dataset.get_split_targets(split="test")
        report.generate_report(test_targets, predictions)



from dlomix.constants import ALPHABET_UNMOD
from dlomix.data import RetentionTimeDataset
from dlomix.eval import TimeDeltaMetric

# data path
from dlomix.models import RetentionTimePredictor

DATAPATH = '../example_dataset/proteomTools_train_val.csv'

'''
create dataset:
    - abstracts TF Dataset
    - splits sequences
    - padds sequences to the specified length
    - performs train/val split if needed
'''
d = RetentionTimeDataset(data_source=DATAPATH, sequence_col="sequence", target_col="irt",
                         seq_length=40, normalize_targets=True,
                         batch_size=128, val_ratio=0.2)


'''
create Model:
    - abstracts a RT Prediction Model
    - encoding to integers happens at beginning of the model
    - model has three main parts (embedding, encoder, regressor) with some parametrization
    '''

model = RetentionTimePredictor(embedding_dim=50, seq_length=40, encoder='lstm')


''''

Target is to always have the high-level model functions callabe to abstract out training, but still people
can write their own custom training loops since the model extends tf.keras.Model

 '''

model.compile(optimizer='adam',
              loss='mse',
              metrics=['mean_absolute_error', TimeDeltaMetric()])

history = model.fit(d.train_data, epochs=15, validation_data=d.val_data)

'''

'''


import json
import pandas as pd
import numpy as np
import tensorflow as tf
from dlomix.constants import DEFAULT_PARQUET_ENGINE

"""
 TODO: check if it is better to abstract out a generic class for TF dataset wrapper, including:
 - splitting data logic (e.g. include task-specific stratification based on sequence length, iRT values)
 - loading data logic
 """

# take into consideration if the pandas dataframe is pickled or not and then call read_pickle instead of read_csv
# allow the possiblity to have three different dataset objects, one for train, val, and test


class RetentionTimeDataset:
    r"""A dataset class for Retention Time prediction tasks. It initialize a dataset object wrapping tf.Dataset and some relevant preprocessing steps.

    Parameters
    -----------
    data_source : str, tuple of two numpy.ndarray, numpy.ndarray, optional
        source can be a tuple of two arrays (sequences, targets), single array (sequences), useful for test data, or a str with a file path to a csv file. Defaults to None.
    sep : str, optional
        separator to be used if the data source is a CSV file. Defaults to ",".
    sequence_col :  str, optional
        name of the column containing the sequences in the provided CSV. Defaults to "sequence".
    target_col : str, optional
        name of the column containing the targets (indexed retention time). Defaults to "irt".
    feature_cols : list, optional
        a list of columns containing other features that can be used later as inputs to a model. Defaults to None.
    normalize_targets : bool, optional
        a boolean whether to normalize the targets or not (subtract mean and divied by standard deviation). Defaults to False.
    seq_length : int, optional
        the sequence length to be used, where all sequences will be padded to this length, longer sequences will be removed and not truncated. Defaults to 0.
    batch_size : int, optional
        the batch size to be used for consuming the dataset in training a model. Defaults to 32.
    val_ratio : int, optional
        a fraction to determine the size of the validation data (0.2 = 20%). Defaults to 0.
    seed: int, optional
        a seed to use for splitting the data to allow for a reproducible split. Defaults to 21.
    test :bool, optional
        a boolean whether the dataset is a test dataset or not. Defaults to False.
    path_aminoacid_atomcounts : str, optional
        a string with a path to a CSV table with the atom counts of the different amino acids (can be used for feature extraction). Defaults to None.
    sample_run : bool, optional
        a boolean to limit the number of examples to a small number, SAMPLE_RUN_N, for testing and debugging purposes. Defaults to False.
    """
    ATOM_TABLE = None
    SPLIT_NAMES = ["train", "val", "test"]
    BATCHES_TO_PREFETCH = tf.data.AUTOTUNE

    SAMPLE_RUN_N = 100
    METADATA_KEY = "metadata"
    PARAMS_KEY = "parameters"
    TARGET_NAME_KEY = "target_column_key"

    # TODO: For test dataset --> examples with longer sequences --> do not drop, add NaN for prediction

    def __init__(
        self,
        data_source=None,
        sep=",",
        sequence_col="sequence",
        target_col="irt",
        feature_cols=None,
        normalize_targets=False,
        seq_length=0,
        batch_size=32,
        val_ratio=0,
        seed=21,
        test=False,
        path_aminoacid_atomcounts=None,
        sample_run=False,
    ):
        super(RetentionTimeDataset, self).__init__()

        np.random.seed(seed)
        self.seed = seed

        self.data_source = data_source
        self.sep = sep
        self.sequence_col = sequence_col.lower()
        self.target_col = target_col.lower()
        if feature_cols:
            self.feature_cols = [f.lower() for f in feature_cols]
        else:
            self.feature_cols = []

        self.normalize_targets = normalize_targets
        self.sample_run = sample_run

        # if seq_length is 0 (default) -> no padding
        self.seq_length = seq_length

        self._data_mean = 0
        self._data_std = 1

        self.batch_size = batch_size
        self.val_ratio = val_ratio
        self.testing_mode = test

        self.main_split = (
            RetentionTimeDataset.SPLIT_NAMES[2]
            if self.testing_mode
            else RetentionTimeDataset.SPLIT_NAMES[0]
        )

        self.sequences = None
        self.targets = None
        self.features_df = None
        self.example_id = None

        # if path to counts lookup table is provided, include count features, otherwise not
        self.include_count_features = True if path_aminoacid_atomcounts else False

        if self.include_count_features:
            self.aminoacid_atom_counts_csv_path = (
                path_aminoacid_atomcounts  # "../lookups/aa_comp_rel.csv"
            )
            self._init_atom_table()

        # initialize TF Datasets dict
        self.tf_dataset = (
            {self.main_split: None, RetentionTimeDataset.SPLIT_NAMES[1]: None}
            if val_ratio != 0
            else {self.main_split: None}
        )

        self.indicies_dict = (
            {self.main_split: None, RetentionTimeDataset.SPLIT_NAMES[1]: None}
            if val_ratio != 0
            else {self.main_split: None}
        )

        # if data is provided with the constructor call --> load, otherwise --> done
        if self.data_source is not None:
            self.load_data(data=data_source)

    def _init_atom_table(self):
        atom_counts = pd.read_csv(self.aminoacid_atom_counts_csv_path)
        atom_counts = atom_counts.astype(str)

        keys_tensor = tf.constant(atom_counts["aa"].values)
        values_tensor = tf.constant(
            ["_".join(c) for c in list(atom_counts.iloc[:, 1:].values)]
        )
        init = tf.lookup.KeyValueTensorInitializer(keys_tensor, values_tensor)
        RetentionTimeDataset.ATOM_TABLE = tf.lookup.StaticHashTable(
            init, default_value="0_0_0_0_0"
        )

    def load_data(self, data):
        """Load data into the dataset object, can be used to load data at a later point after initialization.
        This function triggers the whole pipeline of: data loading, validation (against sequence length), splitting, building TensorFlow dataset objects, and apply preprocessing.

        :param data: can be: tuple of two arrays (sequences, targets), single array (sequences), useful for test data, or a `str` with a file path toa csv file
        :return: None
        """
        self.data_source = data

        self._read_data()
        self._validate_remove_long_sequences()
        self._split_data()
        self._build_tf_dataset()
        self._preprocess_tf_dataset()

    """
    numpy array --> either a tuple or a single array
        - Tuple --> means (sequences, targets)
        - single ndarray --> means sequences only, useful for test dataset
    str --> path to csv file or compressed csv file
    """

    def _read_data(self):
        if isinstance(self.data_source, dict):
            self._update_data_loading_for_json_format()

        if isinstance(self.data_source, tuple):
            tuple_size_is_two = len(self.data_source) == 2
            if tuple_size_is_two:
                tuple_elements_are_ndarray = isinstance(
                    self.data_source[0], np.ndarray
                ) and isinstance(self.data_source[1], np.ndarray)
                if tuple_elements_are_ndarray:
                    self.sequences = self.data_source[0]
                    self.targets = self.data_source[1]
            else:
                raise ValueError(
                    "If a tuple is provided, it has to have a length of 2 and both elements should be numpy arrays."
                )

        elif isinstance(self.data_source, np.ndarray):
            self.sequences = self.data_source
            self.targets = np.zeros(self.sequences.shape[0])
            self._data_mean, self._data_std = 0, 1

        elif isinstance(self.data_source, (str, dict)):
            if isinstance(self.data_source, dict):
                #  a dict is passed via the json
                df = pd.DataFrame(self.data_source)
            else:
                # a string path is passed via the json or as a constructor argument
                df = self._resolve_string_data_path()

            # used only for testing with a smaller sample from a csv file
            if self.sample_run:
                df = df.head(RetentionTimeDataset.SAMPLE_RUN_N)

            # lower all column names
            df.columns = [col_name.lower() for col_name in df.columns]

            self.sequences, self.targets = (
                df[self.sequence_col].values,
                df[self.target_col].values,
            )
            self._data_mean, self._data_std = np.mean(self.targets), np.std(
                self.targets
            )

            self.features_df = df[self.feature_cols]
        else:
            raise ValueError(
                "Data source has to be either a tuple of two numpy arrays, a single numpy array, "
                "or a string with a path to a csv/parquet/json file."
            )

        # give the index of the element as an ID for later reference if needed
        self.example_id = list(range(len(self.sequences)))

    def _update_data_loading_for_json_format(self):
        json_dict = self.data_source

        self.data_source = json_dict.get(RetentionTimeDataset.METADATA_KEY, "")
        self.target_col = json_dict.get(RetentionTimeDataset.PARAMS_KEY, {}).get(
            RetentionTimeDataset.TARGET_NAME_KEY, self.target_col
        )
        # ToDo: make dynamic based on parameters
        self.sequence_col = "modified_sequence"

    def _resolve_string_data_path(self):
        is_json_file = self.data_source.endswith(".json")

        if is_json_file:
            json_dict = read_json_file(self.data_source)
            self._update_data_loading_for_json_format(json_dict)

        is_parquet_url = ".parquet" in self.data_source and self.data_source.startswith(
            "http"
        )
        is_parquet_file = self.data_source.endswith(".parquet")
        is_csv_file = self.data_source.endswith(".csv")

        if is_parquet_url or is_parquet_file:
            df = read_parquet_file_pandas(self.data_source, DEFAULT_PARQUET_ENGINE)
            return df
        elif is_csv_file:
            df = pd.read_csv(self.data_source)
            return df
        else:
            raise ValueError(
                "Invalid data source provided as a string, please provide a path to a csv, parquet, or "
                "or a json file."
            )

    def _validate_remove_long_sequences(self) -> None:
        """
        Validate if all sequences are shorter than the padding length, otherwise drop them.
        """
        assert self.sequences.shape[0] > 0, "No sequences in the provided data."
        assert len(self.sequences) == len(
            self.targets
        ), "Count of examples does not match for sequences and targets."

        limit = self.seq_length
        vectorized_len = np.vectorize(lambda x: len(x))
        mask = vectorized_len(self.sequences) <= limit
        self.sequences, self.targets = self.sequences[mask], self.targets[mask]
        # once feature columns are introduced, apply the mask to the feature columns (subset the dataframe as well)

    def _split_data(self):
        n = len(self.sequences)

        if self.val_ratio != 0 and (not self.testing_mode):
            # add randomization for now and later consider the splitting logic
            self.indicies_dict[RetentionTimeDataset.SPLIT_NAMES[1]] = np.arange(n)[
                : int(n * self.val_ratio)
            ]
            self.indicies_dict[self.main_split] = np.arange(n)[
                int(n * self.val_ratio) :
            ]
        else:
            self.indicies_dict[self.main_split] = np.arange(n)

    def _build_tf_dataset(self):
        # consider adding the feature columns or extra inputs
        for split in self.tf_dataset.keys():
            self.tf_dataset[split] = tf.data.Dataset.from_tensor_slices(
                (
                    self.sequences[self.indicies_dict[split]],
                    self.targets[self.indicies_dict[split]],
                )
            )

    def _preprocess_tf_dataset(self):
        for split in self.tf_dataset.keys():
            # avoid normalizing targets for test data --> should not be needed
            if self.normalize_targets and not self.testing_mode:
                self.tf_dataset[split] = self.tf_dataset[split].map(
                    lambda s, t: self._normalize_target(s, t),
                    num_parallel_calls=tf.data.AUTOTUNE,
                )

            self.tf_dataset[split] = (
                self.tf_dataset[split]
                .map(
                    lambda s, t: self._split_sequence(s, t),
                    num_parallel_calls=tf.data.AUTOTUNE,
                )
                .map(
                    lambda s, t: self._pad_sequences(s, t),
                    num_parallel_calls=tf.data.AUTOTUNE,
                )
            )

            if self.include_count_features:
                self.tf_dataset[split] = (
                    self.tf_dataset[split]
                    .map(
                        RetentionTimeDataset._convert_inputs_to_dict,
                        num_parallel_calls=tf.data.AUTOTUNE,
                    )
                    .map(
                        lambda s, t: self._generate_single_counts(s, t),
                        num_parallel_calls=tf.data.AUTOTUNE,
                    )
                    .map(
                        lambda s, t: self._generate_di_counts(s, t),
                        num_parallel_calls=tf.data.AUTOTUNE,
                    )
                )

            self.tf_dataset[split] = (
                self.tf_dataset[split]
                .batch(self.batch_size)
                .prefetch(RetentionTimeDataset.BATCHES_TO_PREFETCH)
            )

    def get_split_targets(self, split="val"):
        """Retrieve all targets (original labels) for a specific split.

        :param split: a string specifiying the split name (train, val, test)
        :return: nd.array with the targets
        """
        if split not in self.indicies_dict.keys():
            raise ValueError(
                "requested split does not exist, availabe splits are: "
                + list(self.indicies_dict.keys())
            )

        return self.targets[self.indicies_dict[split]]

    def denormalize_targets(self, targets):
        """Denormalize the given targets (can also be predictions) by multiplying the standard deviation and adding the mean.

        :param targets: an nd.array with targets or predictions
        :return: a denormalized nd.array with the targets or the predictions
        """
        return targets * self._data_std + self._data_mean

    def _pad_sequences(self, seq, target):
        pad_len = tf.abs(self.seq_length - tf.size(seq))
        paddings = tf.concat([[0], [pad_len]], axis=0)
        seq = tf.pad(seq, [paddings], "CONSTANT")
        seq.set_shape([self.seq_length])
        return seq, target

    def _normalize_target(self, seq, target):

        target = tf.math.divide(
            tf.math.subtract(target, self._data_mean), self._data_std
        )
        return seq, target

    def _split_sequence(self, seq, target):
        splitted_seq = tf.strings.bytes_split(seq)

        return splitted_seq, target

    """
    if more than one input is added, inputs are added to a python dict, the following methods assume that
    """

    @staticmethod
    def _convert_inputs_to_dict(seq, target):
        return {"seq": seq}, target

    def _generate_single_counts(self, inputs, target):
        inputs["counts"] = tf.map_fn(
            lambda x: RetentionTimeDataset.ATOM_TABLE.lookup(x), inputs["seq"]
        )
        inputs["counts"] = tf.map_fn(
            lambda x: tf.strings.split(x, sep="_"), inputs["counts"]
        )
        inputs["counts"] = tf.strings.to_number(inputs["counts"])
        inputs["counts"].set_shape([self.seq_length, 5])

        return inputs, target

    def _generate_di_counts(self, inputs, target):
        # add every two neighboring elements without overlap [0 0 1 1 2 2 .... pad_length/2 pad_length/2]
        segments_to_add = [i // 2 for i in range(self.seq_length)]
        inputs["di_counts"] = tf.math.segment_sum(
            inputs["counts"], tf.constant(segments_to_add)
        )
        inputs["di_counts"].set_shape([self.seq_length // 2, 5])

        return inputs, target

    def _get_tf_dataset(self, split=None):
        assert (
            split in self.tf_dataset.keys()
        ), f"Requested data split is not available, available splits are {self.tf_dataset.keys()}"
        if split in self.tf_dataset.keys():
            return self.tf_dataset[split]
        return self.tf_dataset

    @property
    def train_data(self):
        """TensorFlow Dataset object for the training data"""
        return self._get_tf_dataset(RetentionTimeDataset.SPLIT_NAMES[0])

    @property
    def val_data(self):
        """TensorFlow Dataset object for the validation data"""
        return self._get_tf_dataset(RetentionTimeDataset.SPLIT_NAMES[1])

    @property
    def test_data(self):
        """TensorFlow Dataset object for the test data"""
        return self._get_tf_dataset(RetentionTimeDataset.SPLIT_NAMES[2])

    @property
    def data_mean(self):
        """Mean value of the targets"""
        return self._data_mean

    @property
    def data_std(self):
        """Standard deviation value of the targets"""
        return self._data_std

    @data_mean.setter
    def data_mean(self, value):
        self._data_mean = value

    @data_std.setter
    def data_std(self, value):
        self._data_std = value


# to go to reader classes or reader utils


def read_parquet_file_pandas(filepath, parquet_engine):
    try:
        df = pd.read_parquet(filepath, engine=parquet_engine)
    except ImportError:
        raise ImportError(
            "Parquet engine is missing, please install fastparquet using pip or conda."
        )
    return df


def read_json_file(filepath):
    with open(filepath, "r") as j:
        json_dict = json.loads(j.read())
    return json_dict


if __name__ == "__main__":
    test_data_dict = {
        "metadata": {
            "linear rt": [1, 2, 3],
            "modified_sequence": ["ABC", "ABC", "ABC"],
        },
        "annotations": {},
        "parameters": {"target_column_key": "linear rt"},
    }

    pd.DataFrame(test_data_dict["metadata"]).to_parquet("metadata.parquet")

    test_data_dict_file = {
        "metadata": "metadata.parquet",
        "annotations": {},
        "parameters": {"target_column_key": "linear rt"},
    }

    rtdataset = RetentionTimeDataset(data_source=test_data_dict, seq_length=20)
    print(rtdataset.sequences)
    print(rtdataset.targets)

    rtdataset = RetentionTimeDataset(data_source=test_data_dict_file, seq_length=20)
    print(rtdataset.sequences)
    print(rtdataset.targets)



import numpy as np
import pandas as pd
import tensorflow as tf

from dlomix.utils import convert_nested_list_to_numpy_array

# take into consideration if the pandas dataframe is pickled or not and then call read_pickle instead of read_csv
# allow the possiblity to have three different dataset objects, one for train, val, and test


class IntensityDataset:
    r"""A dataset class for Intensity prediction tasks. It initialize a dataset object wrapping tf.Dataset and some relevant preprocessing steps.

    Parameters
    -----------
    data_source : str, tuple of two numpy.ndarray, numpy.ndarray, optional
        source can be a tuple of two arrays (sequences, targets), single array (sequences), useful for test data, or a str with a file path to a csv file. Defaults to None.
    sep : str, optional
        separator to be used if the data source is a CSV file. Defaults to ",".
    sequence_col :  str, optional
        name of the column containing the sequences in the provided CSV. Defaults to "sequence".
    target_col : str, optional
        name of the column containing the targets (vector of intensities). Defaults to "intensities_raww".
    feature_cols : list, optional
        a list of columns containing other features that can be used later as inputs to a model. Defaults to None.
    normalize_targets : bool, optional
        a boolean whether to normalize the targets or not (subtract mean and divied by standard deviation). Defaults to False.
    seq_length : int, optional
        the sequence length to be used, where all sequences will be padded to this length, longer sequences will be removed and not truncated. Defaults to 0.
    batch_size : int, optional
        the batch size to be used for consuming the dataset in training a model. Defaults to 32.
    val_ratio : int, optional
        a fraction to determine the size of the validation data (0.2 = 20%). Defaults to 0.
    seed: int, optional
        a seed to use for splitting the data to allow for a reproducible split. Defaults to 21.
    test :bool, optional
        a boolean whether the dataset is a test dataset or not. Defaults to False.
    path_aminoacid_atomcounts : str, optional
        a string with a path to a CSV table with the atom counts of the different amino acids (can be used for feature extraction). Defaults to None.
    sample_run : bool, optional
        a boolean to limit the number of examples to a small number, SAMPLE_RUN_N, for testing and debugging purposes. Defaults to False.
    """
    SPLIT_NAMES = ["train", "val", "test"]
    BATCHES_TO_PREFETCH = tf.data.AUTOTUNE

    SAMPLE_RUN_N = 100

    # TODO: For test dataset --> examples with longer sequences --> do not drop, add NaN for prediction

    def __init__(
        self,
        data_source=None,
        sep=",",
        sequence_col="sequence",
        collision_energy_col="collision_energy_aligned_normed",
        precursor_charge_col="precursor_charge_onehot",
        intensities_col="intensities",
        feature_cols=None,
        normalize_targets=False,
        seq_length=0,
        batch_size=32,
        val_ratio=0,
        seed=21,
        test=False,
        path_aminoacid_atomcounts=None,
        sample_run=False,
    ):
        super(IntensityDataset, self).__init__()

        np.random.seed(seed)
        self.seed = seed

        self.data_source = data_source
        self.sep = sep
        self.sequence_col = sequence_col.lower()
        self.collision_energy_col = collision_energy_col.lower()
        self.precursor_charge_col = precursor_charge_col.lower()
        self.intensities_col = intensities_col.lower()

        if feature_cols:
            self.feature_cols = [f.lower() for f in feature_cols]
        else:
            self.feature_cols = []

        self.normalize_targets = normalize_targets
        self.sample_run = sample_run

        # if seq_length is 0 (default) -> no padding
        self.seq_length = seq_length

        self._data_mean = 0
        self._data_std = 1

        self.batch_size = batch_size
        self.val_ratio = val_ratio
        self.testing_mode = test
        self.no_intensities = self.testing_mode

        self.main_split = (
            IntensityDataset.SPLIT_NAMES[2]
            if self.testing_mode
            else IntensityDataset.SPLIT_NAMES[0]
        )

        self.sequences = None
        self.collision_energy = None
        self.precursor_charge = None
        self.intensities = None

        self.features_df = None
        self.example_id = None

        # if path to counts lookup table is provided, include count features, otherwise not
        self.include_count_features = True if path_aminoacid_atomcounts else False

        if self.include_count_features:
            self.aminoacid_atom_counts_csv_path = (
                path_aminoacid_atomcounts  # "../lookups/aa_comp_rel.csv"
            )
            self._init_atom_table()

        # initialize TF Datasets dict
        self.tf_dataset = (
            {self.main_split: None, IntensityDataset.SPLIT_NAMES[1]: None}
            if val_ratio != 0
            else {self.main_split: None}
        )

        self.indicies_dict = (
            {self.main_split: None, IntensityDataset.SPLIT_NAMES[1]: None}
            if val_ratio != 0
            else {self.main_split: None}
        )

        # if data is provided with the constructor call --> load, otherwise --> done
        if self.data_source is not None:
            self.load_data(data=data_source)

    def _init_atom_table(self):
        atom_counts = pd.read_csv(self.aminoacid_atom_counts_csv_path)
        atom_counts = atom_counts.astype(str)

        keys_tensor = tf.constant(atom_counts["aa"].values)
        values_tensor = tf.constant(
            ["_".join(c) for c in list(atom_counts.iloc[:, 1:].values)]
        )
        init = tf.lookup.KeyValueTensorInitializer(keys_tensor, values_tensor)
        IntensityDataset.ATOM_TABLE = tf.lookup.StaticHashTable(
            init, default_value="0_0_0_0_0"
        )

    def load_data(self, data):
        """Load data into the dataset object, can be used to load data at a later point after initialization.
        This function triggers the whole pipeline of: data loading, validation (against sequence length), splitting, building TensorFlow dataset objects, and apply preprocessing.

        :param data: a `str` with a file path to csv file
        :return: None
        """
        self.data_source = data

        self._read_data()
        self._validate_remove_long_sequences()
        self._split_data()
        self._build_tf_dataset()
        self._preprocess_tf_dataset()

    """
    numpy array --> either a tuple or a single array
        - Tuple --> means (sequences, collision_energy, precursor_charge, intensities)
        - single ndarray --> means sequences only, useful for test dataset
    str --> path to csv file or compressed csv file
    """

    def _read_data(self):
        if isinstance(self.data_source, tuple):
            tuple_size_is_three_or_four = (
                len(self.data_source) == 3 or len(self.data_source) == 4
            )
            if tuple_size_is_three_or_four:
                tuple_elements_are_ndarray = all(
                    [isinstance(x, np.ndarray) for x in self.data_source]
                )
                if tuple_elements_are_ndarray:
                    self.sequences = self.data_source[0]
                    self.collision_energy = self.data_source[1]
                    self.precursor_charge = self.data_source[2]
                    if len(self.data_source) == 4:
                        self.intensities = self.data_source[3]
                        self.no_intensities = False

                    else:
                        self.intensities = np.zeros()
                        self.no_intensities = True
            else:
                raise ValueError(
                    "If a tuple is provided, it has to have a length of 4 and all elements should be numpy arrays."
                )
        elif isinstance(self.data_source, str):
            df = pd.read_csv(self.data_source)

            # used only for testing with a smaller sample from a csv file
            if self.sample_run:
                df = df.head(IntensityDataset.SAMPLE_RUN_N)

            # lower all column names
            df.columns = [col_name.lower() for col_name in df.columns]

            # retrieve columns from the dataframe
            self.sequences = df[self.sequence_col]
            self.collision_energy = df[self.collision_energy_col]
            self.precursor_charge = df[self.precursor_charge_col]
            self.intensities = df[self.intensities_col]

            # parse strings into lists, for precursor charge and intensities
            if isinstance(self.precursor_charge.iloc[0], str):
                self.precursor_charge = self.precursor_charge.apply(eval)

            if isinstance(self.intensities.iloc[0], str):
                self.intensities = self.intensities.apply(eval)

            # get numpy arrays with .values() for all inputs and intensities

            self.sequences = self.sequences.values

            # for concatenation later, we expand dimensions
            self.collision_energy = self.collision_energy.values.reshape(-1, 1)

            self.precursor_charge = convert_nested_list_to_numpy_array(
                self.precursor_charge.values, dtype=np.float64
            )
            self.intensities = convert_nested_list_to_numpy_array(
                self.intensities.values
            )

            self.features_df = df[self.feature_cols]
        else:
            raise ValueError(
                "Data source has to be either a tuple of four numpy arrays,"
                "or a string path to a csv file."
            )

        # give the index of the element as an ID for later reference if needed
        self.example_id = list(range(len(self.sequences)))

    def _validate_remove_long_sequences(self) -> None:
        """
        Validate if all sequences are shorter than the padding length, otherwise drop them.
        """
        assert self.sequences.shape[0] > 0, "No sequences in the provided data."

        # check if count of examples matches for all provided inputs
        lengths = [
            len(self.sequences),
            len(self.collision_energy),
            len(self.precursor_charge),
        ]
        if not self.no_intensities:
            lengths = lengths + [len(self.intensities)]

        assert np.all(
            lengths == np.array(lengths[0])
        ), "Count of examples does not match for sequences and targets."

        limit = self.seq_length
        vectorized_len = np.vectorize(lambda x: len(x))
        mask = vectorized_len(self.sequences) <= limit
        self.sequences = self.sequences[mask]
        self.collision_energy = self.collision_energy[mask]
        self.precursor_charge = self.precursor_charge[mask]
        self.intensities = self.intensities[mask]

        # once feature columns are introduced, apply the mask to the feature columns (subset the dataframe as well)

    def _split_data(self):
        n = len(self.sequences)

        if self.val_ratio != 0 and (not self.testing_mode):
            # add randomization for now and later consider the splitting logic
            self.indicies_dict[IntensityDataset.SPLIT_NAMES[1]] = np.arange(n)[
                : int(n * self.val_ratio)
            ]
            self.indicies_dict[self.main_split] = np.arange(n)[
                int(n * self.val_ratio) :
            ]
        else:
            self.indicies_dict[self.main_split] = np.arange(n)

    def _build_tf_dataset(self):
        # consider adding the feature columns or extra inputs
        for split in self.tf_dataset.keys():
            self.tf_dataset[split] = tf.data.Dataset.from_tensor_slices(
                (
                    self.sequences[self.indicies_dict[split]],
                    self.collision_energy[self.indicies_dict[split]],
                    self.precursor_charge[self.indicies_dict[split]],
                    self.intensities[self.indicies_dict[split]],
                )
            )

    def _preprocess_tf_dataset(self):
        # ToDo: convert input to dict and assume this as the general case --> abstract out in parent class

        for split in self.tf_dataset.keys():
            self.tf_dataset[split] = (
                self.tf_dataset[split]
                .map(
                    IntensityDataset._convert_inputs_to_dict,
                    num_parallel_calls=tf.data.AUTOTUNE,
                )
                .map(
                    lambda i, t: self._split_sequence(i, t),
                    num_parallel_calls=tf.data.AUTOTUNE,
                )
                .map(
                    lambda i, t: self._pad_sequences(i, t),
                    num_parallel_calls=tf.data.AUTOTUNE,
                )
            )

            # Here: feature engineering on the fly if needed (atom counts, etc...)

            self.tf_dataset[split] = (
                self.tf_dataset[split]
                .batch(self.batch_size)
                .prefetch(IntensityDataset.BATCHES_TO_PREFETCH)
            )

    def get_split_targets(self, split="val"):
        """Retrieve all targets (original labels) for a specific split.

        :param split: a string specifiying the split name (train, val, test)
        :return: nd.array with the targets
        """
        if split not in self.indicies_dict.keys():
            raise ValueError(
                "requested split does not exist, availabe splits are: "
                + list(self.indicies_dict.keys())
            )

        return self.intensities[self.indicies_dict[split]]

    def denormalize_targets(self, targets):
        """Denormalize the given targets (can also be predictions) by multiplying the standard deviation and adding the mean.

        :param targets: an nd.array with targets or predictions
        :return: a denormalized nd.array with the targets or the predictions
        """
        return targets * self._data_std + self._data_mean

    def _pad_sequences(self, inputs, target):
        pad_len = tf.abs(self.seq_length - tf.size(inputs["sequence"]))
        paddings = tf.concat([[0], [pad_len]], axis=0)
        inputs["sequence"] = tf.pad(inputs["sequence"], [paddings], "CONSTANT")
        inputs["sequence"].set_shape([self.seq_length])
        return inputs, target

    def _normalize_target(self, seq, target):

        target = tf.math.divide(
            tf.math.subtract(target, self._data_mean), self._data_std
        )
        return seq, target

    def _split_sequence(self, inputs, target):

        inputs["sequence"] = tf.strings.bytes_split(inputs["sequence"])

        return inputs, target

    """
    if more than one input is added, inputs are added to a python dict, the following methods assume that
    """

    @staticmethod
    def _convert_inputs_to_dict(seq, collision, precursor, target):
        inputs_dict = {
            "sequence": seq,
            "collision_energy": collision,
            "precursor_charge": precursor,
        }
        return inputs_dict, target

    def _generate_single_counts(self, inputs, target):
        inputs["counts"] = tf.map_fn(
            lambda x: IntensityDataset.ATOM_TABLE.lookup(x), inputs["seq"]
        )
        inputs["counts"] = tf.map_fn(
            lambda x: tf.strings.split(x, sep="_"), inputs["counts"]
        )
        inputs["counts"] = tf.strings.to_number(inputs["counts"])
        inputs["counts"].set_shape([self.seq_length, 5])

        return inputs, target

    def _generate_di_counts(self, inputs, target):
        # add every two neighboring elements without overlap [0 0 1 1 2 2 .... pad_length/2 pad_length/2]
        segments_to_add = [i // 2 for i in range(self.seq_length)]
        inputs["di_counts"] = tf.math.segment_sum(
            inputs["counts"], tf.constant(segments_to_add)
        )
        inputs["di_counts"].set_shape([self.seq_length // 2, 5])

        return inputs, target

    def _get_tf_dataset(self, split=None):
        assert (
            split in self.tf_dataset.keys()
        ), f"Requested data split is not available, available splits are {self.tf_dataset.keys()}"
        if split in self.tf_dataset.keys():
            return self.tf_dataset[split]
        return self.tf_dataset

    @property
    def train_data(self):
        """TensorFlow Dataset object for the training data"""
        return self._get_tf_dataset(IntensityDataset.SPLIT_NAMES[0])

    @property
    def val_data(self):
        """TensorFlow Dataset object for the validation data"""
        return self._get_tf_dataset(IntensityDataset.SPLIT_NAMES[1])

    @property
    def test_data(self):
        """TensorFlow Dataset object for the test data"""
        return self._get_tf_dataset(IntensityDataset.SPLIT_NAMES[2])

    @property
    def data_mean(self):
        """Mean value of the targets"""
        return self._data_mean

    @property
    def data_std(self):
        """Standard deviation value of the targets"""
        return self._data_std

    @data_mean.setter
    def data_mean(self, value):
        self._data_mean = value

    @data_std.setter
    def data_std(self, value):
        self._data_std = value



import pickle

import tensorflow as tf

from dlomix.data.RetentionTimeDataset import RetentionTimeDataset
from dlomix.eval.rt_eval import TimeDeltaMetric
from dlomix.models.deepLC import DeepLCRetentionTimePredictor

model = DeepLCRetentionTimePredictor(seq_length=50)
opt = tf.keras.optimizers.SGD(learning_rate=1e-4, momentum=0.9)

model.compile(optimizer=opt,
              loss='mse',
              metrics=['mean_absolute_error', TimeDeltaMetric()])


DATAPATH = '../example_dataset/proteomTools_train_val.csv'

d = RetentionTimeDataset(data_source=DATAPATH, seq_length=50, batch_size=512, val_ratio=0.2,
                         path_aminoacid_atomcounts="./lookups/aa_comp_rel.csv")

history = model.fit(d.tf_dataset['train'], epochs=3, validation_data=d.tf_dataset['val'])

model.save_weights('./output/deeplc_test')

with open("./history_deeplc.pkl", 'wb') as f:
   pickle.dump(history.history, f)





import pickle
import numpy as np


def save_obj(obj, name):
    with open(name + ".pkl", "wb") as f:
        pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)


def load_obj(name):
    with open(name + ".pkl", "rb") as f:
        return pickle.load(f)


def convert_nested_list_to_numpy_array(nested_list, dtype=np.float32):
    return np.array([np.array(x, dtype=dtype) for x in nested_list])



import tensorflow as tf
import tensorflow.keras.backend as K


class TimeDeltaMetric(tf.keras.metrics.Metric):
    """Implementation of the time delta metric as a Keras Metric.

    Parameters
    ----------
        mean (int, optional): Mean value of the targets in case normalization was performed. Defaults to 0.
        std (int, optional): Standard deviation value of the targets in case normalization was performed. Defaults to 1.
        percentage (float, optional): What percentage of the data points to consider, this is specific to the conmputation of the metric. Defaults to 0.95 which corresponds to 95% of the datapoints and is the mostly used value in papers.
        name (str, optional): Name of the metric so that it can be reported and used later in Keras History objects. Defaults to 'timedelta'.
        rescale_targets (bool, optional): Whether to rescale (denormalize) targets or not. Defaults to False.
        rescale_predictions (bool, optional): Whether to rescale (denormalize) predictions or not. Defaults to False.
        double_delta (bool, optional): Whether to multiple the computed delta by 2 in order to make it two-sided or not. Defaults to False.
    """

    def __init__(
        self,
        mean=0,
        std=1,
        percentage=0.95,
        name="timedelta",
        rescale_targets=False,
        rescale_predictions=False,
        double_delta=False,
        **kwargs
    ):

        super(TimeDeltaMetric, self).__init__(name=name, **kwargs)
        self.delta = self.add_weight(name="delta", initializer="zeros")
        self.batch_count = self.add_weight(name="batch-count", initializer="zeros")
        self.mean = mean
        self.std = std
        self.percentage = percentage
        self.rescale_targets = rescale_targets
        self.rescale_predictions = rescale_predictions
        self.double_delta = double_delta

    def update_state(self, y_true, y_pred, sample_weight=None):
        # rescale
        if self.rescale_targets:
            y_true = y_true * self.std + self.mean

        if self.rescale_predictions:
            y_pred = y_pred * self.std + self.mean

        # find position of the index
        length = tf.shape(y_true)[0]
        mark = tf.cast(length, dtype=tf.float32) * self.percentage
        mark = tf.cast(mark, dtype=tf.int32)

        # compute residuals and sort
        abs_error = tf.abs(y_true - y_pred)
        d = tf.sort(abs_error)[mark - 1]

        # two-sided delta
        if self.double_delta:
            d = d * 2

        # update count of batches
        self.batch_count.assign_add(1.0)

        # update delta
        self.delta.assign_add(tf.math.reduce_sum(d))

    def result(self):
        # this is simple averaging over the batches, more complex reduction can be added based on domain expertises
        # Examples are: take max or min of both deltas (translates to a strict or a relaxed metric)
        return tf.math.divide(self.delta, self.batch_count)


# code adopted and modified based on:
# https://github.com/horsepurve/DeepRTplus/blob/cde829ef4bd8b38a216d668cf79757c07133b34b/RTdata_emb.py
def delta95_metric(y_true, y_pred):
    mark95 = tf.cast(
        tf.cast(tf.shape(y_true)[0], dtype=tf.float32) * 0.95, dtype=tf.int32
    )
    abs_error = K.abs(y_true - y_pred)
    delta = tf.sort(abs_error)[mark95 - 1]
    norm_range = K.max(y_true) - K.min(y_true)
    return (delta * 2) / (norm_range)



from os.path import join

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from matplotlib.colors import LogNorm
from matplotlib.ticker import LogLocator

from .postprocessing import normalize_intensity_predictions
from .Report import PDFFile, Report


class IntensityReport(Report):
    """Report generation for Fragment Ion Intensity Prediction tasks."""

    TARGETS_LABEL = "x"
    PREDICTIONS_LABEL = "y"
    DEFAULT_BATCH_SIZE = 600

    def __init__(self, output_path, history, figures_ext="png", batch_size=0):
        super(IntensityReport, self).__init__(output_path, history, figures_ext)

        self.pdf_file = PDFFile("DLOmix - Fragment Ion Intensity Report")

        if batch_size:
            self.batch_size = batch_size
        else:
            self.batch_size = IntensityReport.DEFAULT_BATCH_SIZE

    def generate_report(self, dataset, predictions):
        self._init_report_resources()

        predictions_df = self.generate_intensity_results_df(dataset, predictions)
        self.plot_all_metrics()

        # make custom plots
        self.plot_spectral_angle(predictions_df)
        
        self._compile_report_resources_add_pdf_pages()
        self.pdf_file.output(join(self._output_path, "intensity_Report.pdf"), "F")


    def generate_intensity_results_df(self, dataset, predictions):
        predictions_df = pd.DataFrame()

        predictions_df['sequences'] = dataset.sequences
        predictions_df['intensities_pred'] = predictions.tolist()
        predictions_df['precursor_charge_onehot'] = dataset.precursor_charge.tolist()
        predictions_df['intensities_raw'] = dataset.intensities.tolist()

        return predictions_df

    def plot_spectral_angle(
        self,
        predictions_df
        ):
        """Create spectral  plot

        Arguments
        ---------
            predictions_df:  dataframe with raw intensities, predictions, sequences, precursor_charges
        """
        
        predictions_acc = normalize_intensity_predictions(predictions_df, self.batch_size)
        violin_plot = sns.violinplot(predictions_acc['spectral_angle'])
        
        save_path = join(self._output_path, "violin_spectral_angle_plot" + self._figures_ext)

        fig = violin_plot.get_figure()
        fig.savefig(save_path)

        self._add_report_resource(
            "spectral_angle_plot",
            "Spectral angle plot",
            "The following figure shows the spectral angle plot for the test data.",
            save_path,
        )



import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K


def masked_spectral_distance(y_true, y_pred):
    """Masked, normalized spectral angles between true and pred vectors
    > arccos(1*1 + 0*0) = 0         > SL = 0    > high correlation
    > arccos(0*1 + 1*0) = pi/2      > SL = 1    > low correlation
    """

    # To avoid numerical instability during training on GPUs,
    # we add a fuzzing constant epsilon of 1×10−7 to all vectors
    epsilon = K.epsilon()

    # Masking: we multiply values by (true + 1) because then the peaks that cannot
    # be there (and have value of -1 as explained above) won't be considered
    pred_masked = ((y_true + 1) * y_pred) / (y_true + 1 + epsilon)
    true_masked = ((y_true + 1) * y_true) / (y_true + 1 + epsilon)

    # L2 norm
    pred_norm = K.l2_normalize(true_masked, axis=-1)
    true_norm = K.l2_normalize(pred_masked, axis=-1)

    # Spectral Angle (SA) calculation
    # (from the definition below, it is clear that ions with higher intensities
    #  will always have a higher contribution)
    product = K.sum(pred_norm * true_norm, axis=1)
    arccos = tf.math.acos(product)
    return 2 * arccos / np.pi


def masked_pearson_correlation_distance(y_true, y_pred):
    epsilon = K.epsilon()

    # Masking: we multiply values by (true + 1) because then the peaks that cannot
    # be there (and have value of -1 as explained above) won't be considered
    pred_masked = ((y_true + 1) * y_pred) / (y_true + 1 + epsilon)
    true_masked = ((y_true + 1) * y_true) / (y_true + 1 + epsilon)

    mx = tf.math.reduce_mean(true_masked)
    my = tf.math.reduce_mean(pred_masked)
    xm, ym = true_masked-mx, pred_masked-my
    r_num = tf.math.reduce_mean(tf.multiply(xm, ym))
    r_den = tf.math.reduce_std(xm) * tf.math.reduce_std(ym)
    return 1 - (r_num/r_den)



import tensorflow as tf
from dlomix.constants import ALPHABET_UNMOD
from tensorflow.keras.layers.experimental import preprocessing


class DeepLCRetentionTimePredictor(tf.keras.Model):
    def __init__(
        self, seq_length=60, vocab_dict=ALPHABET_UNMOD, use_global_features=False
    ):
        super(DeepLCRetentionTimePredictor, self).__init__()
        self.seq_length = seq_length
        self._use_global_features = use_global_features

        self.leaky_relu = tf.keras.layers.ReLU(max_value=20, negative_slope=0.1)
        self.string_lookup = preprocessing.StringLookup(
            vocabulary=list(vocab_dict.keys())
        )

        self._build_aminoacid_branch()
        self._build_diaminoacid_branch()
        self._build_onehot_encoding_branch()
        self._build_regressor()
        self.output_layer = tf.keras.layers.Dense(1)

        if self._use_global_features:
            self._build_global_features_branch()

    def _build_aminoacid_branch(self):
        self.aminoacid_branch = tf.keras.Sequential(
            [
                self._build_conv_pool_block(n_filters=256, kernel=8, padding="same"),
                self._build_conv_pool_block(n_filters=128, kernel=8, padding="same"),
                self._build_conv_pool_block(
                    n_filters=64, kernel=8, padding="same", pool=False
                ),
                tf.keras.layers.Flatten(),
            ]
        )

    def _build_diaminoacid_branch(self):

        self.diaminoacid_branch = tf.keras.Sequential(
            [
                self._build_conv_pool_block(n_filters=128, kernel=2, padding="same"),
                self._build_conv_pool_block(n_filters=64, kernel=2, padding="same"),
                tf.keras.layers.Flatten(),
            ]
        )

    def _build_global_features_branch(self):
        self.global_features_branch = tf.keras.Sequential(
            [
                tf.keras.layers.Dense(16, activation=self.leaky_relu),
                tf.keras.layers.Dense(16, activation=self.leaky_relu),
                tf.keras.layers.Dense(16, activation=self.leaky_relu),
            ]
        )

    def _build_onehot_encoding_branch(self):
        self.onehot_encoding_branch = tf.keras.Sequential(
            [
                self._build_conv_pool_block(
                    n_filters=2,
                    kernel=2,
                    padding="same",
                    activation="tanh",
                    pool_strides=10,
                    pool_size=10,
                ),
                tf.keras.layers.Flatten(),
            ]
        )

    def _build_regressor(self):
        self.regressor = tf.keras.Sequential(
            [tf.keras.layers.Dense(128, activation=self.leaky_relu) for _ in range(5)]
        )

    def _build_conv_pool_block(
        self,
        n_conv_layers=2,
        n_filters=256,
        kernel=8,
        padding="same",
        activation="leaky_relu",
        pool=True,
        pool_strides=2,
        pool_size=2,
    ):

        # leaky relu by default
        activation_fn = self.leaky_relu

        if activation in ["tanh", "relu"]:
            activation_fn = activation

        block = tf.keras.Sequential(
            [
                tf.keras.layers.Conv1D(
                    filters=n_filters,
                    kernel_size=kernel,
                    padding=padding,
                    activation=activation_fn,
                )
                for _ in range(n_conv_layers)
            ]
        )
        if pool:
            block.add(tf.keras.layers.MaxPooling1D(pool_size=2, strides=2))

        return block

    def call(self, inputs, **kwargs):
        outputs = {}

        integer_encoded = self.string_lookup(inputs["seq"])
        onehot_encoded = tf.one_hot(integer_encoded, depth=self.seq_length)

        if self._use_global_features:
            outputs["global_features_output"] = self.global_features_branch(
                inputs["global_features"]
            )

        outputs["onehot_branch_output"] = self.onehot_encoding_branch(onehot_encoded)
        outputs["aminoacids_branch_output"] = self.aminoacid_branch(inputs["counts"])
        outputs["diaminoacids_branch_output"] = self.diaminoacid_branch(
            inputs["di_counts"]
        )

        concatenated_output = tf.concat(outputs.values(), axis=1)
        concatenated_output = self.regressor(concatenated_output)
        return self.output_layer(concatenated_output)



__version__ = "0.0.4"

META_DATA = {
    "author": "Omar Shouman",
    "author_email": "o.shouman@tum.de",
    "description": "Deep Learning for Proteomics",
    "package_name": "DLOmix",
    "copyright_text": "2022, Wilhelm Lab, TU Munich.",
    "github_url": "https://github.com/wilhelm-lab/dlomix",
}



import tensorflow as tf
from tensorflow.keras.layers.experimental import preprocessing
from dlomix.constants import ALPHABET_UNMOD
from dlomix.layers.attention import AttentionLayer, DecoderAttentionLayer


class PrositRetentionTimePredictor(tf.keras.Model):
    """Implementation of the Prosit model for retention time prediction.

    Parameters
    -----------
        embedding_output_dim (int, optional): Size of the embeddings to use. Defaults to 16.
        seq_length (int, optional): Sequence length of the peptide sequences. Defaults to 30.
        vocab_dict (dict, optional): Dictionary mapping for the vocabulary (the amino acids in this case). Defaults to ALPHABET_UNMOD.
        dropout_rate (float, optional): Probability to use for dropout layers in the encoder. Defaults to 0.5.
        latent_dropout_rate (float, optional): Probability to use for dropout layers in the regressor layers after encoding. Defaults to 0.1.
        recurrent_layers_sizes (tuple, optional): A tuple of 2 values for the sizes of the two GRU layers in the encoder. Defaults to (256, 512).
        regressor_layer_size (int, optional): Size of the dense layer in the regressor after the encoder. Defaults to 512.
    """

    def __init__(
        self,
        embedding_output_dim=16,
        seq_length=30,
        vocab_dict=ALPHABET_UNMOD,
        dropout_rate=0.5,
        latent_dropout_rate=0.1,
        recurrent_layers_sizes=(256, 512),
        regressor_layer_size=512,
    ):
        super(PrositRetentionTimePredictor, self).__init__()

        # tie the count of embeddings to the size of the vocabulary (count of amino acids)
        self.embeddings_count = len(vocab_dict) + 2

        self.dropout_rate = dropout_rate
        self.latent_dropout_rate = latent_dropout_rate
        self.regressor_layer_size = regressor_layer_size
        self.recurrent_layers_sizes = recurrent_layers_sizes

        self.string_lookup = preprocessing.StringLookup(
            vocabulary=list(vocab_dict.keys())
        )

        self.embedding = tf.keras.layers.Embedding(
            input_dim=self.embeddings_count,
            output_dim=embedding_output_dim,
            input_length=seq_length,
        )
        self._build_encoder()

        self.attention = AttentionLayer()

        self.regressor = tf.keras.Sequential(
            [
                tf.keras.layers.Dense(self.regressor_layer_size, activation="relu"),
                tf.keras.layers.Dropout(rate=self.latent_dropout_rate),
            ]
        )

        self.output_layer = tf.keras.layers.Dense(1)

    def _build_encoder(self):
        self.encoder = tf.keras.Sequential(
            [
                tf.keras.layers.Bidirectional(
                    tf.keras.layers.GRU(
                        units=self.recurrent_layers_sizes[0], return_sequences=True
                    )
                ),
                tf.keras.layers.Dropout(rate=self.dropout_rate),
                tf.keras.layers.GRU(
                    units=self.recurrent_layers_sizes[1], return_sequences=True
                ),
                tf.keras.layers.Dropout(rate=self.dropout_rate),
            ]
        )

    def call(self, inputs, **kwargs):
        x = self.string_lookup(inputs)
        x = self.embedding(x)
        x = self.encoder(x)
        x = self.attention(x)
        x = self.regressor(x)
        x = self.output_layer(x)
        return x


class PrositIntensityPredictor(tf.keras.Model):
    """Implementation of the Prosit model for intensity prediction.

    Parameters
    -----------
        embedding_output_dim (int, optional): Size of the embeddings to use. Defaults to 16.
        seq_length (int, optional): Sequence length of the peptide sequences. Defaults to 30.
        vocab_dict (dict, optional): Dictionary mapping for the vocabulary (the amino acids in this case). Defaults to ALPHABET_UNMOD.
        dropout_rate (float, optional): Probability to use for dropout layers in the encoder. Defaults to 0.5.
        latent_dropout_rate (float, optional): Probability to use for dropout layers in the regressor layers after encoding. Defaults to 0.1.
        recurrent_layers_sizes (tuple, optional): A tuple of 2 values for the sizes of the two GRU layers in the encoder. Defaults to (256, 512).
        regressor_layer_size (int, optional): Size of the dense layer in the regressor after the encoder. Defaults to 512.
    """

    def __init__(
        self,
        embedding_output_dim=16,
        seq_length=30,
        len_fion=6,
        vocab_dict=ALPHABET_UNMOD,
        dropout_rate=0.2,
        latent_dropout_rate=0.1,
        recurrent_layers_sizes=(256, 512),
        regressor_layer_size=512,
    ):
        super(PrositIntensityPredictor, self).__init__()

        # tie the count of embeddings to the size of the vocabulary (count of amino acids)
        self.embeddings_count = len(vocab_dict) + 2

        self.dropout_rate = dropout_rate
        self.latent_dropout_rate = latent_dropout_rate
        self.regressor_layer_size = regressor_layer_size
        self.recurrent_layers_sizes = recurrent_layers_sizes
        self.seq_length = seq_length
        self.len_fion = len_fion

        # maximum number of fragment ions
        self.max_ion = self.seq_length - 1

        self.string_lookup = preprocessing.StringLookup(
            vocabulary=list(vocab_dict.keys())
        )

        self.embedding = tf.keras.layers.Embedding(
            input_dim=self.embeddings_count,
            output_dim=embedding_output_dim,
            input_length=seq_length,
        )

        self._build_encoders()
        self._build_decoder()

        self.attention = AttentionLayer(name="encoder_att")

        self.fusion_layer = tf.keras.Sequential(
            [
                tf.keras.layers.Multiply(name="add_meta"),
                tf.keras.layers.RepeatVector(self.max_ion, name="repeat"),
            ]
        )

        self.regressor = tf.keras.Sequential(
            [
                tf.keras.layers.TimeDistributed(
                    tf.keras.layers.Dense(self.len_fion), name="time_dense"
                ),
                tf.keras.layers.LeakyReLU(name="activation"),
                tf.keras.layers.Flatten(name="out"),
            ]
        )

    def _build_encoders(self):

        self.meta_encoder = tf.keras.Sequential(
            [
                tf.keras.layers.Concatenate(name="meta_in"),
                tf.keras.layers.Dense(
                    self.recurrent_layers_sizes[1], name="meta_dense"
                ),
                tf.keras.layers.Dropout(self.dropout_rate, name="meta_dense_do"),
            ]
        )

        self.sequence_encoder = tf.keras.Sequential(
            [
                tf.keras.layers.Bidirectional(
                    tf.keras.layers.GRU(
                        units=self.recurrent_layers_sizes[0], return_sequences=True
                    )
                ),
                tf.keras.layers.Dropout(rate=self.dropout_rate),
                tf.keras.layers.GRU(
                    units=self.recurrent_layers_sizes[1], return_sequences=True
                ),
                tf.keras.layers.Dropout(rate=self.dropout_rate),
            ]
        )

    def _build_decoder(self):
        self.decoder = tf.keras.Sequential(
            [
                tf.keras.layers.GRU(
                    units=self.regressor_layer_size,
                    return_sequences=True,
                    name="decoder",
                ),
                tf.keras.layers.Dropout(rate=self.dropout_rate),
                DecoderAttentionLayer(self.max_ion),
            ]
        )

    def call(self, inputs, **kwargs):
        peptides_in = inputs["sequence"]
        collision_energy_in = inputs["collision_energy"]
        precursor_charge_in = inputs["precursor_charge"]

        encoded_meta = self.meta_encoder([collision_energy_in, precursor_charge_in])

        x = self.string_lookup(peptides_in)
        x = self.embedding(x)
        x = self.sequence_encoder(x)
        x = self.attention(x)

        x = self.fusion_layer([x, encoded_meta])

        x = self.decoder(x)

        x = self.regressor(x)

        return x



import functools

import numpy as np
import tensorflow as tf

import dlomix.losses as losses


def reshape_dims(array):
    n, dims = array.shape
    assert dims == 174
    nlosses = 1
    return array.reshape(
        [array.shape[0], 30 - 1, 2, nlosses, 3]
    )


def reshape_flat(array):
    s = array.shape
    flat_dim = [s[0], functools.reduce(lambda x, y: x * y, s[1:], 1)]
    return array.reshape(flat_dim)


def normalize_base_peak(array):
    # flat
    maxima = array.max(axis=1)
    array = array / maxima[:, np.newaxis]
    return array


def mask_outofrange(array, lengths, mask=-1.):
    # dim
    for i in range(array.shape[0]):
        array[i, lengths[i] - 1 :, :, :, :] = mask
    return array


def mask_outofcharge(array, charges, mask=-1.):
    # dim
    for i in range(array.shape[0]):
        if charges[i] < 3:
            array[i, :, :, :, charges[i] :] = mask
    return array


def get_spectral_angle(true, pred, batch_size=600):

    n = true.shape[0]
    sa = np.zeros([n])

    def iterate():
        if n > batch_size:
            for i in range(n // batch_size):
                true_sample = true[i * batch_size : (i + 1) * batch_size]
                pred_sample = pred[i * batch_size : (i + 1) * batch_size]
                yield i, true_sample, pred_sample
            i = n // batch_size
            yield i, true[(i) * batch_size :], pred[(i) * batch_size :]
        else:
            yield 0, true, pred

    for i, t_b, p_b in iterate():
        tf.compat.v1.reset_default_graph()
        with tf.compat.v1.Session() as s:
            sa_graph = losses.masked_spectral_distance(t_b, p_b)
            sa_b = 1 - s.run(sa_graph)
            sa[i * batch_size : i * batch_size + sa_b.shape[0]] = sa_b
    sa = np.nan_to_num(sa)
    return sa


def normalize_intensity_predictions(data, batch_size=600):
    assert "sequences" in data, "Key sequences is missing in the data provided for post-processing"
    assert "intensities_pred" in data, "Key intensities_pred is missing in the data provided for post-processing"
    assert "precursor_charge_onehot" in data, "Key precursor_charge_onehot is missing in the data provided for post-processing"

    sequence_lengths = data["sequences"].apply(lambda x: len(x))
    intensities =  np.stack(data["intensities_pred"].to_numpy()).astype(np.float32)
    precursor_charge_onehot = np.stack(data["precursor_charge_onehot"].to_numpy())
    charges = list(precursor_charge_onehot.argmax(axis=1) + 1)

    intensities[intensities < 0] = 0
    intensities = reshape_dims(intensities)
    intensities = mask_outofrange(intensities, sequence_lengths)
    intensities = mask_outofcharge(intensities, charges)
    intensities = reshape_flat(intensities)
    m_idx = intensities == -1
    intensities = normalize_base_peak(intensities)
    intensities[m_idx] = -1
    data["intensities_pred"] = intensities

    if "intensities_raw" in data:
        data["spectral_angle"] = get_spectral_angle(
            np.stack(data["intensities_raw"].to_numpy()).astype(np.float32), intensities, batch_size=batch_size
        )
    return data


import tensorflow as tf
from tensorflow.keras.layers.experimental import preprocessing
from dlomix.constants import ALPHABET_UNMOD


class RetentionTimePredictor(tf.keras.Model):
    """A simple class for Retention Time prediction models.

    Parameters
    ----------
        embedding_dim (int, optional): Dimensionality of the embeddings to be used for representing the Amino Acids. Defaults to 16.
        seq_length (int, optional): Sequence length of the peptide sequences. Defaults to 30.
        encoder (str, optional): String for specifying the decoder to use, either based on 1D conv-layers or LSTMs. Defaults to "conv1d".
        vocab_dict (dict, optional): Dictionary mapping for the vocabulary (the amino acids in this case). Defaults to ALPHABET_UNMOD.
    """

    def __init__(
        self,
        embedding_dim=16,
        seq_length=30,
        encoder="conv1d",
        vocab_dict=ALPHABET_UNMOD,
    ):
        super(RetentionTimePredictor, self).__init__()

        # tie the count of embeddings to the size of the vocabulary (count of amino acids)
        self.embeddings_count = len(vocab_dict) + 2

        self.string_lookup = preprocessing.StringLookup(
            vocabulary=list(vocab_dict.keys())
        )

        self.embedding = tf.keras.layers.Embedding(
            input_dim=self.embeddings_count,
            output_dim=embedding_dim,
            input_length=seq_length,
        )

        self._build_encoder(encoder)

        self.flatten = tf.keras.layers.Flatten()
        self.regressor = tf.keras.Sequential(
            [
                tf.keras.layers.Dense(128, activation="relu"),
                tf.keras.layers.Dense(64, activation="relu"),
            ]
        )

        self.output_layer = tf.keras.layers.Dense(1)

    def _build_encoder(self, encoder_type):
        if encoder_type.lower() == "conv1d":
            self.encoder = tf.keras.Sequential(
                [
                    tf.keras.layers.Conv1D(
                        filters=256, kernel_size=3, padding="same", activation="relu"
                    ),
                    tf.keras.layers.Conv1D(
                        filters=512, kernel_size=3, padding="valid", activation="relu"
                    ),
                    tf.keras.layers.MaxPooling1D(pool_size=2),
                ]
            )
        else:
            self.encoder = tf.keras.Sequential(
                [
                    tf.keras.layers.LSTM(256, return_sequences=True),
                    tf.keras.layers.LSTM(256),
                ]
            )

    def call(self, inputs, **kwargs):
        x = self.string_lookup(inputs)
        x = self.embedding(x)
        x = self.encoder(x)
        x = self.flatten(x)
        x = self.regressor(x)
        x = self.output_layer(x)

        return x



from dlomix.data import RetentionTimeDataset, IntensityDataset
import numpy as np
import pandas as pd


INTENSITY_CSV_EXAMPLE_URL = 'https://raw.githubusercontent.com/wilhelm-lab/dlomix/develop/example_dataset/intensity/intensity_data.csv'
TEST_DATA_PARQUET = "metadata.parquet"

def test_empty_rtdataset():
    rtdataset = RetentionTimeDataset()
    assert rtdataset.sequences is None
    assert rtdataset.targets is None
    assert rtdataset.main_split is RetentionTimeDataset.SPLIT_NAMES[0]

def test_simple_rtdataset():
    rtdataset = RetentionTimeDataset(data_source=(np.array(['AAA', 'BBB']), np.array([21.5, 26.5])))
    assert rtdataset.sequences is not None
    assert rtdataset.targets is not None
    assert rtdataset.main_split is RetentionTimeDataset.SPLIT_NAMES[0]

def test_json_dict_rtdataset():
    test_data_dict = {
        "metadata": {
            "linear rt" : [1,2,3],
            "modified_sequence": ["ABC", "ABC", "ABC"]
        },
        "annotations": {},
        "parameters": {
            "target_column_key": "linear rt"
        }
    }

    pd.DataFrame(test_data_dict["metadata"]).to_parquet(TEST_DATA_PARQUET)

    test_data_dict_file = {
        "metadata": TEST_DATA_PARQUET,
        "annotations": {},
        "parameters": {
            "target_column_key": "linear rt"
        }
    }

    rtdataset = RetentionTimeDataset(data_source=test_data_dict, seq_length=20)
    rtdataset_filebased = RetentionTimeDataset(data_source=test_data_dict_file, seq_length=20)

def test_empty_intensitydataset():
    intensity_dataset = IntensityDataset()
    assert intensity_dataset.sequences is None
    assert intensity_dataset.collision_energy is None
    assert intensity_dataset.precursor_charge is None
    assert intensity_dataset.intensities is None
    assert intensity_dataset.main_split is IntensityDataset.SPLIT_NAMES[0]

def test_simple_intensitydataset():
    intensity_dataset = IntensityDataset(data_source=(
        np.array(['SVFLTFLR']),
        np.array([0.25]),
        np.array([[0, 1, 0, 0, 0, 0]]),
        np.array([[0.03713018032121684, 0.0, -1.0, 0.0, 0.0, -1.0, 0.02485036943573326, 0.0, -1.0, 0.37425569938350733, 0.0, -1.0, 0.1006487907071137, 0.0, -1.0, 0.16793299234113923, 0.0, -1.0, 0.5770605328948204, 0.004866043683849902, -1.0, 0.013969858753800551, 0.0, -1.0, 0.3613063752966507, 0.004158167348899733, -1.0, 0.004756058682204546, 0.0, -1.0, 1.0, 0.05804204277785504, -1.0, 0.0, 0.0, -1.0, 0.0026942297891076857, 0.0042070812435137245, -1.0, 0.0, 0.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0]])
    ))

    assert intensity_dataset.sequences is not None
    assert intensity_dataset.collision_energy is not None
    assert intensity_dataset.precursor_charge is not None
    assert intensity_dataset.intensities is not None
    assert intensity_dataset.main_split is IntensityDataset.SPLIT_NAMES[0]

def test_csv_intensitydataset():
    intensity_dataset = IntensityDataset(data_source=INTENSITY_CSV_EXAMPLE_URL)
    
    assert intensity_dataset.sequences is not None
    assert intensity_dataset.collision_energy is not None
    assert intensity_dataset.precursor_charge is not None
    assert intensity_dataset.intensities is not None
    assert intensity_dataset.main_split is IntensityDataset.SPLIT_NAMES[0]




import os
import sys
import pickle
import tensorflow as tf
import pandas as pd

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from dlomix.eval import TimeDeltaMetric
from dlomix.models import PrositRetentionTimePredictor
from dlomix.data import RetentionTimeDataset
from dlomix.reports import RetentionTimeReport

# consider the use-case for starting from a saved model

model = PrositRetentionTimePredictor(seq_length=30)

optimizer = tf.keras.optimizers.Adam(lr=0.0001, decay=1e-7)

TRAIN_DATAPATH = '../example_dataset/proteomTools_train_val.csv'
TEST_DATAPATH = '../example_dataset/proteomTools_test.csv'

d = RetentionTimeDataset(data_source=TRAIN_DATAPATH, seq_length=30, batch_size=512, val_ratio=0.2)

model.compile(optimizer=optimizer,
              loss='mse',
              metrics=['mean_absolute_error', TimeDeltaMetric()])

weights_file = "./prosit_test"
checkpoint = tf.keras.callbacks.ModelCheckpoint(weights_file, save_best_only=True, save_weights_only=True)
decay = tf.keras.callbacks.ReduceLROnPlateau(monitor='val_loss', factor=0.1, patience=10, verbose=1, min_lr=0)
early_stop = tf.keras.callbacks.EarlyStopping(patience=20)
callbacks = [checkpoint, early_stop, decay]


history = model.fit(d.train_data, epochs=150, validation_data=d.val_data, callbacks=callbacks)

test_rtdata = RetentionTimeDataset(data_source=TEST_DATAPATH,
                                   seq_length=30, batch_size=512, test=True)

predictions = model.predict(test_rtdata.test_data)
predictions = d.denormalize_targets(predictions)
predictions = predictions.ravel()
test_targets = test_rtdata.get_split_targets(split="test")

report = RetentionTimeReport(output_path="./output", history=history)

print("R2: ", report.calculate_r2(test_targets, predictions))

pd.DataFrame(
    {"sequence": test_rtdata.sequences, "irt": test_rtdata.targets, "predicted_irt": predictions}
).to_csv("./predictions_prosit_fullrun.csv", index=False)


# TODO: function to store and load history object, or maybe consider saving the report object
with open("./history_prosit.pkl", 'wb') as f:
    pickle.dump(history.history, f)




import os
import sys
import pickle
import tensorflow as tf
import pandas as pd

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))


from dlomix.models import PrositIntensityPredictor
from dlomix.data import IntensityDataset
from dlomix.losses import masked_spectral_distance

# consider the use-case for starting from a saved model

model = PrositIntensityPredictor(seq_length=30)

optimizer = tf.keras.optimizers.Adam(learning_rate=0.0001, decay=1e-7)

TRAIN_DATAPATH = '../example_dataset/intensity/intensity_data.csv'
#TEST_DATAPATH = '../example_dataset/proteomTools_test.csv'

d = IntensityDataset(data_source=TRAIN_DATAPATH, seq_length=30, batch_size=128, val_ratio=0.3)

# continue here: test dataset ???

model.compile(optimizer=optimizer,
              loss=masked_spectral_distance,
              metrics=['mse'])

weights_file = "./prosit_intensity_test"
checkpoint = tf.keras.callbacks.ModelCheckpoint(weights_file, save_best_only=True, save_weights_only=True)
decay = tf.keras.callbacks.ReduceLROnPlateau(monitor='val_loss', factor=0.1, patience=10, verbose=1, min_lr=0)
early_stop = tf.keras.callbacks.EarlyStopping(patience=20)
callbacks = [checkpoint, early_stop, decay]


history = model.fit(d.train_data, epochs=5, validation_data=d.val_data, callbacks=callbacks)


predictions = model.predict(d.val_data)
# predictions = d.denormalize_targets(predictions)

print(predictions.shape)
print(predictions[0])



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

from dlomix import META_DATA, __version__

VERSION = __version__

setuptools.setup(
    name=META_DATA['package_name'].lower(),
    version=VERSION,
    author=META_DATA['author'],
    author_email=META_DATA['author_email'],
    description=META_DATA['description'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=META_DATA['github_url'],
    packages=setuptools.find_packages(),
    install_requires=[
        'fpdf',
        'pandas',
        'numpy',
        'matplotlib',
        'scikit-learn',
        'tensorflow',
        'pyarrow',
        'seaborn',
        ],
    extras_require={
        "dev": [
            "pytest >= 3.7",
            "pytest-cov",
            "black",
            "twine",
            "setuptools",
            "wheel",
            "pylint",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Development Status :: 1 - Planning",
        "Intended Audience :: Science/Research"
    ],
)



from fpdf import FPDF
from matplotlib import pyplot as plt
from os.path import join
from os import makedirs
import tensorflow as tf
import warnings
import abc
import glob


class Report(abc.ABC):
    """Base class for reports, child classes should implement the abstract method generate_report.

    Parameters
    ----------
        output_path: path to save output files and figures.
        history : reference to a Keras History object or its history dict attribute (History.history).
        figures_ext: File extension and format for saving figures.
    """

    VALID_FIGURE_FORMATS = ["pdf", "jpeg", "jpg", "png"]

    def __init__(self, output_path, history, figures_ext):
        self._output_path = output_path
        makedirs(self._output_path, exist_ok=True)

        if history is None:
            warnings.warn(
                "The passed History object is None, no training/validation data can be reported."
            )
            self._history_dict = {}
        else:
            self._set_history_dict(history)
        self._set_figures_format(figures_ext)

        # an empty dict to use to list the report resources
        self._init_report_resources()

    def _set_history_dict(self, history):
        if isinstance(history, dict):
            self._history_dict = history
        elif not isinstance(history, tf.keras.callbacks.History):
            raise ValueError(
                "Reporting requires a History object (tf.keras.callbacks.History) or its history dict attribute (History.history), which is returned from a call to "
                "model.fit(). Passed history argument is of type {} ",
                type(history),
            )
        elif not hasattr(history, "history"):
            raise ValueError(
                "The passed History object does not have a history attribute, which is a dict with results."
            )
        else:
            self._history_dict = history.history

        if len(self._history_dict.keys()) == 0:
            warnings.warn(
                "The passed History object contains an empty history dict, no training was done."
            )

    def _set_figures_format(self, figures_ext):
        figures_ext = figures_ext.lower()
        if figures_ext.startswith("."):
            figures_ext = figures_ext[1:]
        if figures_ext not in Report.VALID_FIGURE_FORMATS:
            raise ValueError(
                "Allowed figure formats are: {}", Report.VALID_FIGURE_FORMATS
            )
        self._figures_ext = "." + figures_ext

    def _get_all_saved_plots(self):
        all_plots = glob.glob(join(self._output_path, "*" + self._figures_ext))
        return all_plots

    def _add_report_resource(self, key, title, paragraph_text, value):
        self._report_resources[key] = (title, paragraph_text, value)

    def _init_report_resources(self):
        self._report_resources = {}

    def _compile_report_resources_add_pdf_pages(self):
        for key, resource in self._report_resources.items():
            value_is_fig_path = self._figures_ext in str(resource[2])
            plot_word_is_in_key = "plot" in key
            if value_is_fig_path or plot_word_is_in_key:
                self.pdf_file.add_content_plot_page(
                    section_title=resource[0],
                    section_body=resource[1],
                    plot_filepath=resource[2],
                )
            else:
                self.pdf_file.add_content_text_page(
                    section_title=resource[0], section_body=resource[1]
                )

    def plot_keras_metric(self, metric_name, save_plot=True):
        """Plot a keras metric given its name and the history object returned by model.fit()

        Arguments
        ---------
            metric_name: String with the name of the metric.
            save_plot (bool, optional): whether to save plot to disk or not. Defaults to True.
        """

        if metric_name.lower() not in self._history_dict.keys():
            raise ValueError(
                "Metric name to plot is not available in the history dict. Available metrics to plot are {}",
                self._history_dict.keys(),
            )

        if "val_" + metric_name.lower() not in self._history_dict.keys():
            raise ValueError(
                "No validation epochs were run during training, the metric name to plot is not available in the history dict. Available metrics to plot are {}",
                self._history_dict.keys(),
            )
        plt.plot(self._history_dict[metric_name])
        plt.plot(self._history_dict["val_" + metric_name])
        plt.title(metric_name)
        plt.ylabel(metric_name)
        plt.xlabel("epoch")
        plt.legend(["train", "val"], loc="upper left")
        if save_plot:
            save_path = join(self._output_path, metric_name + self._figures_ext)
            plt.savefig(save_path)
        plt.show()
        plt.close()
        metric_name_spaced = metric_name.replace("_", " ")
        self._add_report_resource(
            metric_name + "_plot",
            metric_name_spaced.title(),
            f"The following figure shows the {metric_name_spaced} for training and validation.",
            save_path,
        )

    def plot_all_metrics(self):
        """Plot all available Keras metrics in the History object."""
        metrics = self._history_dict.keys()
        metrics = filter(lambda x: not x.startswith(tuple(["val_", "_"])), metrics)
        for metric in metrics:
            self.plot_keras_metric(metric)

    @abc.abstractmethod
    def generate_report(self, targets, predictions, **kwargs):
        """Abstract method to generate a complete report. Child classes need to implement this method.

        Arguments
        ---------
            targets: Array with target values.
            predictions: Array with prediction values.
        """
        pass


class PDFFile(FPDF):
    """PDF file template class.

    Parameters
    ----------
        title: Title for the pdf file
    """

    PAGE_WIDTH = 210
    PAGE_HEIGHT = 297

    SECTION_PARAGRAPH_FONT = ["Arial", "", 11]
    SECTION_TITLE_FONT = ["Arial", "B", 13]
    LINE_HEIGHT = 5

    def __init__(self, title):

        super().__init__()
        self.title = title
        self.width = PDFFile.PAGE_WIDTH
        self.height = PDFFile.PAGE_HEIGHT

        self.set_auto_page_break(True)
        self.document_empty = True

    def header(self):
        self.set_font("Arial", "B", 11)
        self.cell(self.width - 80)
        self.cell(60, 1, self.title, 0, 0, "R")
        self.ln(20)

    def footer(self):
        # Page numbers in the footer
        self.set_y(-15)
        self.set_font("Arial", "I", 8)
        self.set_text_color(128)
        self.cell(0, 10, "Page " + str(self.page_no()), 0, 0, "C")

    def _add_plot(self, plot_filepath):
        self.image(plot_filepath)
        self.ln(3 * PDFFile.LINE_HEIGHT)

    def _add_section_content(self, section_title, section_body):
        if section_title != "":
            self.set_font(*PDFFile.SECTION_TITLE_FONT)
            self.cell(w=0, txt=section_title)
            self.ln(PDFFile.LINE_HEIGHT)
        if section_body != "":
            self.set_font(*PDFFile.SECTION_PARAGRAPH_FONT)
            self.multi_cell(w=0, h=PDFFile.LINE_HEIGHT, txt=section_body)
            self.ln(PDFFile.LINE_HEIGHT)

    def _create_first_page_if_document_empty(self):
        if self.document_empty:
            self.add_page()
            self.document_empty = False

    def add_content_text_page(self, section_title, section_body):
        """Add a section title and a paragraph.

        Arguments
        ---------
            section_title: title for the section.
            section_body: paragraph text to add.
        """
        self._create_first_page_if_document_empty()
        self._add_section_content(section_title, section_body)

    def add_content_plot_page(self, plot_filepath, section_title="", section_body=""):
        """Add a new page with a section title, a paragraph, and a plot. At least a plot has to be provided.

        Arguments
        ---------
            plot_filepath (str): filepath of the plot to be inserted in the new page.
            section_title (str, optional): title for the section. Defaults to "".
            section_body (str, optional): paragraph text to add. Defaults to "".
        """

        self._create_first_page_if_document_empty()
        self._add_section_content(section_title, section_body)
        self._add_plot(plot_filepath)




from dlomix.models import PrositRetentionTimePredictor, PrositIntensityPredictor
import logging

logger = logging.getLogger(__name__)



def test_prosit_retention_time_model():
    model = PrositRetentionTimePredictor()
    logger.info(model)
    assert model is not None

def test_prosit_intensity_model():
    model = PrositIntensityPredictor()
    logger.info(model)
    assert model is not None



